<?php

namespace AltSolution\Admin;

use AltSolution\Admin\Acl;
use AltSolution\Admin\Form;
use AltSolution\Admin\Http\Controllers;
use AltSolution\Admin\Http\View\Composers;
use AltSolution\Admin\System;
use Illuminate\Contracts\Auth\Access\Gate;
use Illuminate\Foundation\Application;
use Illuminate\Routing\Router;
use Illuminate\Support\ServiceProvider;
use Illuminate\Contracts\Events\Dispatcher as DispatcherContract;
use Illuminate\Contracts\Debug\ExceptionHandler as ExceptionHandlerContract;
use Illuminate\Support\Facades\Event;

class AdminServiceProvider extends ServiceProvider
{
    public function register()
    {
        $this->mergeConfigFrom(__DIR__ .'/../config/admin.php', 'admin');

        $this->registerSystem();
        $modules = $this->registerModules();
        $this->registerContracts();

        $this->app->register(Providers\SeoProvider::class);
        $this->app->register(Providers\UserProvider::class);
        if (isset($modules['admin.module.menu'])) {
            $this->app->register(Providers\MenuProvider::class);
        }
        if (isset($modules['admin.module.content'])) {
            $this->app->register(Providers\ContentProvider::class);
        }
        if (isset($modules['admin.module.cron'])) {
            $this->app->register(Providers\CronProvider::class);
        }
    }

    /**
     * Bootstrap the application services.
     *
     * @return void
     */
    public function boot()
    {
        $this->bootRoutes();
        $this->bootViews();

        $this->loadTranslationsFrom(__DIR__ . '/../resources/lang', 'admin');

        $this->bootPolicies();
        if ($this->app->runningInConsole()) {
            $this->bootPublishes();
        }

        // users last login handler
        Event::listen(\Illuminate\Auth\Events\Login::class, Listeners\AuthLogin::class);
        // log users auth events
        Event::subscribe(Listeners\AuthLogSubscriber::class);

//        /** @var DispatcherContract $events */
//        $events = $this->app->make(DispatcherContract::class);
//        $events->listen();
//        $events->subscribe();
    }

    private function bootRoutes()
    {
        $router = $this->app->make(Router::class);

        $router->middlewareGroup('admin', [
            // TODO: get rid of "web"
            'web',
            \Illuminate\Auth\Middleware\Authenticate::class,
            Http\Middleware\BootSystem::class,
        ]);

        if (!$this->app->routesAreCached()) {
            /** @var Router $router */

            // MAIN CYCLE:
            //[request]->
            //  [middleware]admin->
            //      [router]->[controller]->[action]permission:*

            $prefix = cms_admin_url();

            // assets routes
            $router->get('/vendor/admin/{path}', [Controllers\AssetController::class, 'index'])->where('path', '.+');

            // auth routes
            $router->middleware('web')
                ->prefix($prefix)
                ->as('admin::')
                ->group(function (Router $router) {
                    // log in/out
                    $router->get('login', [Controllers\Auth\LoginController::class, 'login'])->name('login');
                    $router->post('login', [Controllers\Auth\LoginController::class, 'postLogin'])->name('login_fulfil');
                    $router->get('logout', [Controllers\Auth\LoginController::class, 'logout'])->name('logout');

                    if ($this->app->make('config')->get('admin.auth_allow_password_reset')) {
                        // password reset
                        $router->get('password/forgot', [Controllers\Auth\PasswordController::class, 'forgot'])->name('password_forgot');
                        $router->post('password/forgot', [Controllers\Auth\PasswordController::class, 'postForgot'])->name('password_forgot_fulfil');
                        $router->get('password/reset/{token}', [Controllers\Auth\PasswordController::class, 'reset'])->name('password_link');
                        $router->post('password/reset', [Controllers\Auth\PasswordController::class, 'postReset'])->name('password_reset');
                    }

                    // 2FA
                    $router->get('login/verify', [Controllers\Auth\F2AController::class, 'verify'])->name('2fa');
                    $router->post('login/verify', [Controllers\Auth\F2AController::class, 'verifyPost'])->name('2fa_fulfil');

                    // Impersonate
                    $router->get('impersonate/start/{id}', [Controllers\Auth\ImpersonateController::class, 'start'])->name('impersonate');
                    $router->get('impersonate/stop', [Controllers\Auth\ImpersonateController::class, 'stop'])->name('impersonate_stop');
                });

            /** @var System\ModuleInterface[] $modules */
            $modules = $this->app->tagged('admin.system.module');
            $router->middleware('admin')
                ->prefix($prefix)
                ->as('admin::')
                ->group(function (Router $router) use ($modules) {
                    foreach ($modules as $module) {
                        // TODO: routes registered on any request
                        $module->registerRoutes($router);
                    }
                });
        }
    }

    private function bootViews()
    {
        $this->loadViewsFrom(__DIR__ . '/../resources/views', 'admin');
        view()->composer('admin::layouts.auth', Composers\AuthLayoutComposer::class);
        view()->composer('admin::layouts.default', Composers\LayoutComposer::class);
    }

    private function bootPolicies()
    {
        /** @var Gate $gate */
        $gate = $this->app->make(Gate::class);
        /** @var Acl\AclRepositoryInterface $aclRepo */
        $aclRepo = $this->app->make(Acl\AclRepositoryInterface::class);
        $gate->define('permission', [$aclRepo, 'userHasPermission']);
    }

    private function bootPublishes()
    {
        $this->publishes([
            __DIR__ . '/../resources/views' => resource_path('views/vendor/admin'),
        ], 'admin-views');

        $this->publishes([
            __DIR__ . '/../resources/lang' => resource_path('lang/vendor/admin'),
        ], 'admin-lang');

        $this->publishes([
            __DIR__ . '/../assets' => public_path('vendor/admin'),
        ], 'admin-assets');

        $this->publishes([
            __DIR__ . '/../database/migrations/' => database_path('migrations')
        ], 'admin-migrations');

        // $this->publishes([
        //     $configFile => config_path('<some>.php'),
        // ], 'admin-config');
    }

    private function registerSystem()
    {
        $this->app->singleton(System\System::class);
        $this->app->singleton(System\AuthLayout::class);
        $this->app->singleton(System\Layout::class);

        $this->app->afterResolving(Controllers\Auth\Controller::class, function (Controllers\Auth\Controller $controller, Application $app) {
            $controller->setLayout($app->make(System\AuthLayout::class));
            $controller->boot();
        });
        $this->app->afterResolving(Controllers\Controller::class, function (Controllers\Controller $controller, Application $app) {
            $controller->setLayout($app->make(System\Layout::class));
            $controller->boot();
        });

        // admin exception handler
        $previousHandler = null;
        if ($this->app->bound(ExceptionHandlerContract::class) === true) {
            $previousHandler = $this->app->make(ExceptionHandlerContract::class);
        }
        $this->app->singleton(ExceptionHandlerContract::class, function () use ($previousHandler) {
            return new \AltSolution\Admin\Exceptions\Handler($previousHandler);
        });
    }

    private function registerModules()
    {
        $modules = $this->app['config']['admin.modules'];
        $disabled = $this->app['config']['admin.modules_disabled'];
        foreach ($disabled as $alias) {
            unset($modules[$alias]);
        }
        foreach ($modules as $alias => $class) {
            $this->app->bind($alias, $class);
        }
        $this->app->tag(array_keys($modules), 'admin.system.module');

        return $modules;
    }

    private function registerContracts()
    {
        // forms
        $this->app->bind(Form\BuilderInterface::class, Form\Builder::class);
        //$this->app->singleton(System\AssetInterface::class, System\Asset::class);

        if (!$this->app->configurationIsCached()) {
            // twig forms
            $this->app['config']->prepend('twigbridge.extensions.enabled', Twig\Extension\FormExtension::class);
            // twig cms
            $this->app['config']->prepend('twigbridge.extensions.enabled', Twig\Extension\HelpersExtension::class);
        }

        // options
        $this->app->singleton(Option\RepositoryInterface::class, Option\Repository::class);

        // auth
        $this->app->bind(Auth\UserStoreInterface::class, Auth\SessionUserStore::class);
        if ($this->app['config']['admin.auth_2fa_google']) {
            $this->app->bind(Auth\QRCodeInterface::class, Auth\GoogleQRCode::class);
            $this->app->bind(Auth\TwoFactorAuthInterface::class, Auth\GoogleTFA::class);
        } else {
            $this->app->bind(Auth\TwoFactorAuthInterface::class, Auth\CodeTFA::class);
            $this->app->bind(Auth\TwoFactorCodeInterface::class, Auth\MailTFC::class);
        }
        // TODO:
        //$this->app->alias(Auth\TwoFactorAuthInterface::class, 'admin.auth.2fa');
        //$this->app->alias(Auth\TwoFactorCodeInterface::class, 'admin.auth.2fa.code');

        // acl
        $this->app->singleton(Acl\AclRepositoryInterface::class, Acl\AclRepository::class);

        // email templates
        $this->app->singleton(EmailTemplate\MailerInterface::class, EmailTemplate\Mailer::class);
        $this->app->singleton(EmailTemplate\TemplateRepository::class);
        $this->app->singleton(EmailTemplate\RenderInterface::class, EmailTemplate\TwigRender::class);

        // for ImagesTrait
        $this->app->singleton(Image\ImageManagerInterface::class, function() {
            $config = $this->app->make('config');
            // TODO: move to ci
            $options = [
                'directory' => $config['admin.upload_directory'],
                'url' => $config['admin.upload_url'],
                'permission' => $config['admin.upload_permission'],
                'keep_source' => $config['admin.upload_keep_source'],
                'double_encode' => $config['admin.upload_double_encode'],
                'slice_upload' => $config['admin.upload_slice_upload'],
                'slice_verbose' => $config['admin.upload_slice_verbose'],
            ];
            return new Image\ImageManager($options);
        });

        // for UploadsTrait
        $this->app->singleton(Upload\UploadManagerInterface::class, function() {
            $config = $this->app->make('config');
            // TODO: move to ci
            $options = [
                'directory' => $config['admin.upload_directory'],
                'url' => $config['admin.upload_url'],
                'permission' => $config['admin.upload_permission'],
            ];
            return new Upload\UploadManager($options);
        });
    }

}
